"""
AutoPLEX -- Automated machine-learned Potential Landscape explorer -- flows.

This package is designed to perform a workflow-based
automated machine-learned interatomic potentials (MLIP) fit for
DFT-labelled data.

"""
